# -*- coding: utf-8 -*-
from __future__ import absolute_import, division

import os
import sys
from threading import Thread, Lock

from Crypto.Cipher import AES
from emokit.util import crypto_key
from queue import Queue


class EmotivCrypto:
    def __init__(self, serial_number=None, is_research=False):
        """
        Performs decryption of packets received. Stores decrypted packets in a Queue for use.

        :param serial_number - The serial number to use for AES key generation.
        :param is_research - Is this a research edition headset? Also, EPOC+ uses this now.
        """
        # Where the encrypted data is Queued.
        self._encrypted_queue = Queue()
        # Where the decrypted data is Queued.
        self._decrypted_queue = Queue()
        # Running state.
        self.running = False
        # Stop signal tells the loop to stop after processing remaining tasks.
        self._stop_signal = False
        # The emotiv serial number.
        self.serial_number = serial_number
        # EPOC+ and research edition may need to have this set to True
        # TODO: Add functions that check variance in data received. If extreme toggle is_research and check again.
        #  If extreme again raise exception and shutdown emokit.
        self.is_research = is_research
        self.lock = Lock()
        self.thread = Thread(target=self.run)
        self.thread.setDaemon(True)

    def run(self):
        """
        The crypto loop. Decrypts data in encrypted Queue and puts it onto the decrypted Queue.

        Do not call explicitly, use .start() instead.
        """
        # Initialize AES
        cipher = self.new_cipher()
        self.lock.acquire()
        while self.running:
            self.lock.release()
            # While the encrypted queue is not empty.
            while not self._encrypted_queue.empty():
                # Get some encrypted data off of the encrypted Queue.
                encrypted_data = self._encrypted_queue.get()
                # Make sure the encrypted data is not None.
                if encrypted_data is not None:
                    # Make sure the encrypted data is not empty.
                    if len(encrypted_data):
                        try:
                            # Python 3 compatibility
                            if sys.version_info >= (3, 0):
                                # Convert to byte array or bytes like object.
                                encrypted_data = bytes(encrypted_data, encoding='latin-1')
                            # Decrypt the encrypted data.
                            decrypted_data = decrypt_data(cipher, encrypted_data)
                            # Put the decrypted data onto the decrypted Queue.
                            self._decrypted_queue.put_nowait(decrypted_data)
                        except Exception as ex:
                            # Catch everything, and print exception.
                            # TODO: Make this more specific perhaps?
                            print("Emotiv CryptoError ", sys.exc_info()[0], sys.exc_info()[1], sys.exc_info()[2], " : ",
                                  ex)
            # If stop signal is received and all the pending data in the encrypted Queue is processed, stop running.
            self.lock.acquire()
            if self._stop_signal and self._encrypted_queue.empty():
                print("Crypto stopping.")
                self.running = False
        self.lock.release()

    def start(self):
        """
        Starts the crypto thread.
        """
        self.running = True
        self.thread.start()

    def stop(self):
        """
        Stops the crypto thread.
        """
        self.lock.acquire()
        self._stop_signal = True
        self.lock.release()
        self.thread.join(60)

    def new_cipher(self):
        """
        Generates a new AES cipher from the serial number and headset version.
        :return: New AES cipher
        """
        # Create initialization vector.
        iv = os.urandom(AES.block_size)
        # Make sure the serial number was set.
        if self.serial_number is None:
            raise ValueError("Serial number must not be None.")
        # Create and return new AES class, using the serial number and headset version.
        return AES.new(crypto_key(self.serial_number, self.is_research), AES.MODE_ECB, iv)

    def add_task(self, data):
        """
        Gives the crypto thread some encrypted data to decrypt, unless the crypto class' _stop_signal is True.
        :param data: Encrypted Data
        """
        # If the stop signal has not been set yet.
        if not self._stop_signal:
            # Add encrypted data to the encrypted Queue.
            self._encrypted_queue.put_nowait(data)

    def get_data(self):
        """
        Gives decrypted data from the crypto thread, if the queue isn't empty.
        :return: Decrypted data or None
        """
        # If the decrypted queue is not empty, get data from the Queue and return it.
        if not self._decrypted_queue.empty():
            return self._decrypted_queue.get_nowait()
        # Otherwise, return None.
        return None

    def data_ready(self):
        """
        :return: If queue is not empty, return True
        """
        if not self._decrypted_queue.empty():
            return True
        return False

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
            Do cleanup stuff.
        :param exc_type:
        :param exc_val:
        :param exc_tb:
        :return:
        """
        self.stop()


def decrypt_data(cipher, data):
    """
    Returns decrypted data.
    :param cipher: AES cipher
    :param data: Encrypted Data
    :return: Decrypted Data
    """
    return cipher.decrypt(data[:16]) + cipher.decrypt(data[16:])



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division

import os
import platform
import sys
from datetime import datetime
from threading import Thread, Lock

from emokit.decrypter import EmotivCrypto
from emokit.reader import EmotivReader
from queue import Queue

from .packet import EmotivPacket
from .sensors import sensors_mapping
from .util import EmotivWriter

system_platform = platform.system()


class Emotiv(object):
    """
    Receives, decrypts and stores packets received from Emotiv Headsets and other sources.
    """

    # TODO: Add a calibration mechanism, get a "noise average" per sensor or even a "noise pattern" to filter
    #       sensor values when processing packets received. Ideally this would be done not on someone's head.
    # TODO: Add filters for facial expressions, muscle contractions.
    def __init__(self, display_output=False, serial_number=None, is_research=False, write=False,
                 write_encrypted=False, write_values=True, input_source="emotiv",
                 sys_platform=system_platform, verbose=False):
        """
        Sets up initial values.

        :param display_output - Should non-error output be displayed to console?
        :param serial_number - Specify serial_number, needed to decrypt packets for raw data reader and special cases.
        :param is_research - Is EPOC headset research edition? Doesn't seem to work even if it is.
        :param write - Write data to csv.
        :param write_encrypted - Write encrypted data
        :param write_values - Write decrypted sensor data, True overwrites exporting data from dongle pre-processing.
        :param input_source - Source to read from, emotiv or a file name
                   (must be a csv, exported from emokit or following our format)
        :param sys_platform - Operating system, to avoid global statement

        Expect performance to suffer when writing data to a csv.

        """
        self.running = False
        # Queue with EmotivPackets that have been received.
        self.packets = Queue()
        # Battery percent, as int.
        self.battery = 0
        # Display sensor values in console.
        self.display_output = display_output
        # Print details of operation in console, exceptions, etc.
        self.verbose = verbose
        # Should be True for research edition EPOCs and newer EPOC+s, apparently.
        self.is_research = is_research
        self.sensors = sensors_mapping
        # The EPOCs serial number.
        self.serial_number = serial_number
        # Only really applies to quality value calculation, probably needs some updating.
        self.old_model = False
        # Write data to disk.
        self.write = write
        # Assume we are going to read encrypted data from the headset.
        self.read_encrypted = True
        # Not used at the moment.
        self.read_values = False
        # Write the data received as is.
        self.write_encrypted = False
        # Write the data received after decrypting it.
        self.write_decrypted = False
        # Write values and quality values of each sensor.
        self.write_values = write_values
        # The current platform.
        self.platform = sys_platform
        # The number of times data was received.
        self.packets_received = 0
        # The number of times data was decrypted or made into EmotivPackets.
        self.packets_processed = 0
        # The source of the emotiv data.
        self.input_source = input_source
        # If the input source is not an emotiv headset, it is a file.
        self.reader = None
        self.lock = Lock()
        # Setup output writers, multiple types can be output now at once.
        self.encrypted_writer = None
        self.decrypted_writer = None
        self.value_writer = None

        self.crypto = None
        # Setup the crypto thread, if we are reading from an encrypted data source.

        # Setup emokit loop thread. This thread coordinates the work done from the reader to be decrypted and queued
        # into EmotivPackets.
        self.thread = Thread(target=self.run)
        self._stop_signal = False
        self.thread.setDaemon(True)
        self.start()

    def initialize_reader(self):
        if self.input_source != "emotiv":
            # If the name of the input source starts with emotiv_encrypted, get the serial number.
            if self.input_source.startswith('emotiv_encrypted'):
                # Split the name of the input by _'s and retrieve the serial number, typically located after
                # emotiv_encrypted_data_ so index 3.
                self.serial_number = self.input_source.split('_')[3]
            else:
                # Otherwise, check if it is values only and set read_encrypted to False.
                self.read_encrypted = False
                if self.input_source.startswith('emotiv_values'):
                    self.read_values = True
            self.reader = EmotivReader(file_name=self.input_source, mode="csv")
            # Make sure the reader still has the serial number set, since we won't be obtaining it from the headset.
            self.reader.serial_number = self.serial_number
        else:
            # Initialize an EmotivReader with default values, it will try to locate a headset.
            self.reader = EmotivReader()
            if self.reader.serial_number is not None:
                # If EmotivReader found a serial number automatically, change local serial number to the reader serial.
                self.serial_number = self.reader.serial_number

    def initialize_writer(self):
        if self.write_encrypted:
            # Only write encrypted if we are reading encrypted.
            if self.read_encrypted:
                self.encrypted_writer = EmotivWriter('emotiv_encrypted_data_%s_%s.csv' % (
                    self.reader.serial_number, str(datetime.now()).replace(':', '-')), mode="csv")

        # Setup decrypted data writer.
        if self.write_decrypted:
            # If we are reading values we do not have the decrypted data, rather than reconstructing it do not write it.
            if not self.read_values:
                self.decrypted_writer = EmotivWriter('emotiv_data_%s.csv' % str(datetime.now()).replace(':', '-'),
                                                     mode="csv")
        # Setup sensor value writer.
        if self.write_values:
            self.value_writer = EmotivWriter('emotiv_values_%s.csv' % str(datetime.now()).replace(':', '-'),
                                             mode="csv")
            # Make the first row in the file the header with the sensor name
            header_row = []
            for key in self.sensors.keys():
                header_row.append(key + " Value")
                header_row.append(key + " Quality")
            self.value_writer.write(header_row)

    def initialize_crypto(self):
        if self.read_encrypted:
            self.crypto = EmotivCrypto(self.serial_number, self.is_research)

    def start(self):
        """
        Starts emotiv, called upon initialization.
        """
        self.running = True
        self.thread.start()

    def stop(self):
        """
        Stops emotiv
        :return:
        """
        self.reader.stop()
        if self.crypto is not None:
            self.crypto.stop()
        self._stop_signal = True

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if traceback:
            self.log(traceback)
        self.stop()

    def log(self, message):
        """
        Logging function, only prints if verbose is True.
        :param message: Message to log/print
        """
        if self.display_output and self.verbose:
            print("%s" % message)

    def run(self):
        """ Do not call explicitly, called upon initialization of class or self.start()
            The main emokit loop.
        :param reader: EmotivReader class
        :param crypto: EmotivCrypto class
        """
        self.initialize_reader()
        self.initialize_writer()
        self.initialize_crypto()
        self.reader.start()
        if self.crypto is not None:
            self.crypto.start()
        dirty = True
        last_packets_received = 0
        last_packets_decrypted = 0
        tick_time = datetime.now()
        packets_received_since_last_update = 0
        packets_processed_since_last_update = 0
        stale_rx = 0
        restarting_reader = False
        self.lock.acquire()
        while self.running:
            self.lock.release()
            if not self.reader.data.empty():
                try:
                    raw_data = self.reader.data.get()
                except KeyboardInterrupt:
                    self.quit()
                if self.write and self.write_encrypted and self.input_source == 'emotiv':
                    # Due to some encoding problem to save encrypted data we must first convert it to binary.
                    if sys.version_info >= (3, 0):
                        self.encrypted_writer.write(map(bin, bytearray(raw_data, encoding='latin-1')))
                    else:
                        self.encrypted_writer.write(map(bin, bytearray(raw_data)))
                self.packets_received += 1
                if not self.read_encrypted:
                    if not self.read_values:
                        new_packet = EmotivPacket(raw_data)
                        if new_packet.battery is not None:
                            self.battery = new_packet.battery
                        self.packets.put_nowait(new_packet)
                        dirty = True
                        self.packets_processed += 1
                    else:
                        # TODO: Implement read from values.
                        pass
                else:
                    if self.input_source != 'emotiv' and self.read_encrypted:
                        if len(raw_data) != 32:
                            self.reader.lock.acquire()
                            self.reader.running = False
                            self.reader.lock.release()
                            self.crypto.stop()
                            self.running = False
                            raise ValueError("Reached end of data or corrupted data.")
                        # Decode binary data stored in file.

                        if sys.version_info >= (3, 0):
                            raw_data = [int(bytes(item, encoding='latin-1').decode(), 2) for item in raw_data]
                        else:
                            raw_data = [int(item.decode(), 2) for item in raw_data]
                        raw_data = ''.join(map(chr, raw_data[:]))
                    self.crypto.add_task(raw_data)
            if self.crypto is not None:
                if self.crypto.data_ready():
                    decrypted_packet_data = self.crypto.get_data()
                    if self.write_decrypted:
                        if self.decrypted_writer is not None:
                            self.decrypted_writer.write(decrypted_packet_data)
                    self.packets_processed += 1
                    new_packet = EmotivPacket(decrypted_packet_data)
                    if new_packet.battery is not None:
                        self.battery = new_packet.battery
                    self.packets.put_nowait(new_packet)
                    if self.write_values:
                        data_to_write = []
                        for key in self.sensors.keys():
                            data_to_write.extend([new_packet.sensors[key]['value'], new_packet.sensors[key]['quality']])
                        if self.value_writer is not None:
                            self.value_writer.write(data_to_write)
                    dirty = True
            if self._stop_signal:
                self.reader.lock.acquire()
                if not self.reader.running and not restarting_reader:
                    if self.crypto is not None:
                        self.crypto.lock.acquire()
                        if not self.crypto.running and not self.crypto.data_ready():
                            self.running = False
                        self.crypto.lock.release()
                    else:
                        self.running = False
                self.reader.lock.release()

            if tick_time.second != datetime.now().second:
                packets_received_since_last_update = self.packets_received - last_packets_received
                if packets_received_since_last_update == 1 or packets_received_since_last_update == 0:
                    stale_rx += 1
                packets_processed_since_last_update = self.packets_processed - last_packets_decrypted
                last_packets_decrypted = self.packets_processed
                last_packets_received = self.packets_received
                tick_time = datetime.now()
                dirty = True
                self.reader.lock.acquire()
                if restarting_reader and self.reader.stopped:
                    self.reader.lock.release()
                    print("Restarting reader")
                    stale_rx = 0
                    self.initialize_reader()
                    self.reader.start()
                    restarting_reader = False
                    print("End restarting")
                else:
                    self.reader.lock.release()

                if stale_rx > 5 and not restarting_reader:
                    self.reader.stop()
                    restarting_reader = True

            if self.display_output:
                if dirty:
                    if system_platform == "Windows":
                        os.system('cls')
                    else:
                        os.system('clear')
                    print("Packets Received: %s Packets Processed: %s" % (self.packets_received,
                                                                          self.packets_processed))
                    print('\n'.join("%s Reading: %s Quality: %s" %
                                    (k[1], self.sensors[k[1]]['value'],
                                     self.sensors[k[1]]['quality']) for k in enumerate(self.sensors)))
                    print("Battery: %i" % self.battery)
                    print("Sample Rate Rx: {0} Crypto: {1}".format(
                        packets_received_since_last_update,
                        packets_processed_since_last_update)
                    )
                    dirty = False
            self.lock.acquire()
        self.lock.release()

    def dequeue(self):
        """
        Returns an EmotivPacket popped off the Queue.
        """
        try:
            if not self.packets.empty():
                return self.packets.get()
        except KeyboardInterrupt:
            self.quit()
        return None

    def clear_queue(self):
        self.packets = Queue()

    def quit(self):
        """
        A little more forceful stop.
        """
        self.stop()
        os._exit(1)

    def force_quit(self):
        """
        Kill emokit. Might leave files and other objects open or locked.
        """
        os._exit(1)


if __name__ == "__main__":
    a = Emotiv()



# -*- encoding: utf-8 -*-

from emokit.emotiv import Emotiv


def test_emotiv_no_headset():
    try:
        emo_test = Emotiv()
    except Exception as ex:
        assert (ex.message == "Device not found")


def test_import_hidapi():
    hidapi_found = False
    try:
        import hidapi
        hidapi_found = True
    except Exception as ex:
        pass
    try:
        import pywinusb.hid
        hidapi_found = True
    except Exception as ex:
        pass
    assert hidapi_found



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division

import csv
import os
import platform
import sys
import time
from threading import Thread, RLock

from queue import Queue

from .util import validate_data, device_is_emotiv, hid_enumerate, print_hid_enumerate

system_platform = platform.system()
if system_platform == "Windows":
    import pywinusb.hid as hid
else:
    import hidapi


class EmotivReader(object):
    """
    Read data from file or hid. Only CSV for now.
    """

    def __init__(self, file_name=None, mode="hid", hid=None, file=None, **kwargs):
        self.mode = mode
        self.file = file
        self.file_name = file_name
        self.hid = hid
        self.platform = system_platform
        self.serial_number = None
        self.lock = RLock()
        if self.platform != "Windows":
            hidapi.hid_init()
        self.setup_platform = {
            'Windows': self.setup_windows,
            'Darwin': self.setup_not_windows,
            'Linux': self.setup_not_windows,
            'Reader': self.setup_reader,
        }
        if self.mode == "csv":
            if file_name is None:
                raise ValueError("CSV file name must be specified when initializing an EmotivReader class using mode "
                                 "'csv'.")

            if sys.version_info >= (3, 0):
                self.file = open(file_name, 'r')
            else:
                self.file = open(file_name, 'rb')
            self.reader = csv.reader(self.file, quoting=csv.QUOTE_ALL)
            self.platform = "Reader"
        elif self.mode == 'hid':
            self.reader = None
        else:
            self.reader = None
        self.data = Queue()
        self.setup_platform[self.platform]()
        self.running = False
        self.stopped = True
        if self.reader is not None:
            self.thread = Thread(target=self.run, kwargs={'source': self.reader})
        else:
            self.thread = Thread(target=self.run, kwargs={'source': self.hid})
        self.thread.setDaemon(True)
        self._stop_signal = False

    def start(self):
        """
        Starts the reader thread.
        """
        self.running = True
        self.stopped = False
        self.thread.start()

    def stop(self):
        """
        Stops the reader thread.
        """
        self._stop_signal = True

    def run(self, source=None):
        """Do not call explicitly, called upon initialization of class"""
        if self.platform == 'Windows':
            source.set_raw_data_handler(self.data_handler)
            source.open()
        self.lock.acquire()
        while self.running:
            self.lock.release()
            if not self.platform == 'Windows':
                self.lock.acquire()
                try:
                    if not self._stop_signal:
                        data = read_platform[self.platform](source)
                        self.data.put_nowait(data)
                except Exception as ex:
                    print(ex.message)
                    # Catching StopIteration for some reason stops at the second record,
                    #  even though there are more results.
                self.lock.release()
            else:
                time.sleep(0.0005)
            self.lock.acquire()
            if self._stop_signal:
                print("Reader stopping...")
                self.running = False
        self.lock.release()
        if self.file is not None:
            self.file.close()
        if type(source) != int:
            source.close()
        if self.hid is not None:
            if type(self.hid) != int:
                self.hid.close()
        if system_platform != "Windows":
            try:
                hidapi.hid_close(source)
            except Exception:
                pass
            try:
                hidapi.hid_exit()
            except Exception:
                pass
        print("Reader stopped...")
        self.stopped = True
        return

    def data_handler(self, data):
        """
        Receives packets from headset for Windows. Sends them to a Queue to be processed
        by the crypto thread.
        """
        self.lock.acquire()
        if not self._stop_signal:
            data = validate_data(data)
            if data is not None:
                self.data.put_nowait(''.join(map(chr, data[1:])))
        self.lock.release()

    def __exit__(self, exc_type, exc_value, traceback):
        self.stop()
        if self.reader:
            self.reader.close()
        self.file.close()
        if 'eeg_raw' in self.platform and self.hid is not None:
            self.hid.close()
        elif 'Windows' not in self.platform and self.hid is not None:
            hidapi.hid_close(self.hid)

    def setup_reader(self):
        """
        Setup reader stuff, not much to do here right now.
        """
        if 'encrypted' in self.file_name:
            self.platform += ' encrypted'

    def setup_windows(self):
        """
        Setup for headset on the Windows platform.
        """
        devices = []
        try:
            for device in hid.find_all_hid_devices():
                if device_is_emotiv(device, self.platform):
                    devices.append(device)
            if len(devices) == 0:
                print_hid_enumerate(system_platform, hid)
                sys.exit()
            device = devices[1]
            device.open()
            self.hid = device
            self.serial_number = device.serial_number
            device.set_raw_data_handler(self.data_handler)
        except Exception as ex:
            print("Emotiv WindowsSetupError ", sys.exc_info()[0], sys.exc_info()[1], sys.exc_info()[2], " : ", ex)
        finally:
            if self.hid is not None:
                self.hid.close()

    def setup_not_windows(self):
        """
        Setup for headset on a non-windows platform.
        Receives packets from headset and sends them to a Queue to be processed
        by the crypto thread.
        """
        if os.path.exists('/dev/eeg/raw'):
            self.hid = open("/dev/eeg/raw")
        if self.hid is not None:
            # The decryption is handled by the Linux epoc daemon. We don't need to handle it.
            self.platform += " raw_eeg"
        else:
            path, serial_number = hid_enumerate(hidapi, self.platform)
            if len(path) == 0:
                print_hid_enumerate(system_platform, hidapi)
                raise Exception("Device not found")
            self.serial_number = serial_number
            self.hid = hidapi.hid_open_path(path)


def read_csv(source):
    """
    Iterate over data from CSV file.
    :param source: CSV reader
    :return: Next row in CSV file.
    """
    if sys.version_info >= (3, 0):
        return source.__next__()
    else:
        return source.next()


def read_reader(source):
    """
    Read from EmotivReader only. Return data for decryption.
    :param source: Emotiv data reader
    :return: Next row in Emotiv data file.
    """
    data = read_csv(source)
    return data


def read_non_windows(source):
    """
    Read from Emotiv hid device.
    :param source: Emotiv hid device
    :return: Next encrypted packet from Emotiv device.
    """
    # Doesn't seem to matter how big we make the buffer 32 returned every time, 33 for other platforms
    # Set timeout for 1 second, to help with thread shutdown.
    data = validate_data(hidapi.hid_read_timeout(source, 34, 1000))
    if data is not None:
        return ''.join(map(chr, data[1:]))


def read_os_decrypted_non_windows(source):
    """
    Read from Emotiv hid device.
    :param source: Emotiv hid device
    :return: Next packet from Emotiv device.
    """
    data = source.read(32)
    if data != "":
        return data


read_platform = {
    'Darwin': read_non_windows,
    'Linux': read_non_windows,
    'Darwin raw_eeg': read_os_decrypted_non_windows,
    'Linux raw_eeg': read_os_decrypted_non_windows,
    'Reader': read_reader,
    'Reader encrypted': read_reader
}



# -*- coding: utf-8 -*-
import csv
import sys

if sys.version_info >= (3, 0):  # pragma: no cover
    unicode = str


def get_level(data, bits):
    """
    Returns sensor level value from data using sensor bit mask in micro volts (uV).
    """
    level = 0
    for i in range(13, -1, -1):
        level <<= 1
        b = (bits[i] // 8) + 1
        o = bits[i] % 8
        if sys.version_info >= (3, 0):
            level |= (data[b] >> o) & 1
        else:
            level |= (ord(data[b]) >> o) & 1
    return level


def is_old_model(serial_number):
    if "GM" in serial_number[-2:]:
        return False
    return True


def hid_enumerate(hidapi, platform):
    """
    Loops over all devices in the hidapi and attempts to locate the emotiv headset.

    Since hidapi ultimately uses the path there is no reason to get the vendor id or product id.

    Although, they may be useful in locating the device.

    :returns
        path - the path to the device
        serial_number - the serial number of the device
    """
    path = ""
    serial_number = ""
    devices = get_hid_devices[platform](hidapi)
    emotiv_devices = []
    for device in devices:
        if device_is_emotiv(device, platform):
            serial_number = device.serial_number
            path = device.path
            emotiv_devices.append(device)
    if platform != "Windows":
        return path, serial_number
    else:
        return emotiv_devices


def print_hid_device_info_win(device):
    print(unicode(device.vendor_name))
    print(unicode(device.product_name))


def print_hid_device_info_nix(device):
    print(unicode(device.manufacturer_string))
    print(unicode(device.product_string))
    print(device.path)


def print_hid_device_info_all(device):
    print(device.vendor_id)
    print(device.product_id)
    print(device.serial_number)
    print("-------------------------")


def hid_enumerate_nix(hidapi):
    return hidapi.hid_enumerate()


def hid_enumerate_win(hidapi):
    return hidapi.find_all_hid_devices()


def print_hid_enumerate(platform, hidapi):
    """
    Loops over all devices in the hidapi and attempts prints information.

    This is a fall back method that give the user information to give the developers when opening an issue.
    """
    devices = get_hid_devices[platform](hidapi)
    print("-------------------------")
    for device in devices:
        print_device_info[platform](device)
        print_hid_device_info_all(device)
    print("Please include this information if you open a new issue.")


get_hid_devices = {
    'Darwin': hid_enumerate_nix,
    'Linux': hid_enumerate_nix,
    'Windows': hid_enumerate_win
}

print_device_info = {
    'Darwin': print_hid_device_info_nix,
    'Linux': print_hid_device_info_nix,
    'Windows': print_hid_device_info_win
}


def crypto_key(serial_number, is_research=False):
    k = ['\0'] * 16
    k[0] = serial_number[-1]
    k[1] = '\0'
    k[2] = serial_number[-2]
    if is_research:
        k[3] = 'H'
        k[4] = serial_number[-1]
        k[5] = '\0'
        k[6] = serial_number[-2]
        k[7] = 'T'
        k[8] = serial_number[-3]
        k[9] = '\x10'
        k[10] = serial_number[-4]
        k[11] = 'B'
    else:
        k[3] = 'T'
        k[4] = serial_number[-3]
        k[5] = '\x10'
        k[6] = serial_number[-4]
        k[7] = 'B'
        k[8] = serial_number[-1]
        k[9] = '\0'
        k[10] = serial_number[-2]
        k[11] = 'H'
    k[12] = serial_number[-3]
    k[13] = '\0'
    k[14] = serial_number[-4]
    k[15] = 'P'
    return ''.join(k)


def device_is_emotiv(device, platform):
    is_emotiv = False

    try:
        if platform != 'Windows':
            product_name = unicode(device.product_string)
            vendor_name = unicode(device.manufacturer_string)
        else:
            product_name = unicode(device.product_name)
            vendor_name = unicode(device.vendor_name)
        if u"emotiv" in vendor_name.lower():
            is_emotiv = True
        if u"emotiv" in product_name.lower():
            is_emotiv = True
        if u"epoc" in product_name.lower():
            is_emotiv = True
        if u"brain waves" in product_name.lower():
            is_emotiv = True
        if product_name == u'00000000000':
            is_emotiv = True
        if u"eeg signals" in product_name.lower():
            is_emotiv = True
    except Exception as ex:
        print("Emotiv IsEmotivError ", sys.exc_info()[0], sys.exc_info()[1], sys.exc_info()[2], " : ", ex)
    return is_emotiv


def validate_data(data):
    if len(data) == 32:
        data.insert(0, 0)
    if len(data) != 33:
        return None
    return data


class EmotivWriter(object):
    """
    Write data from headset to output. CSV file for now.
    """

    def __init__(self, file_name, mode="csv", **kwargs):
        self.mode = mode
        if sys.version_info >= (3, 0):
            self.file = open(file_name, 'w', newline='')
        else:
            self.file = open(file_name, 'wb')
        if self.mode == "csv":
            self.writer = csv.writer(self.file, quoting=csv.QUOTE_ALL)
        else:
            self.writer = None

    def write_csv(self, data):
        if sys.version_info >= (3, 0):
            if type(data) == str:
                data = bytes(data, encoding='latin-1')
        else:
            if type(data) == str:
                data = [ord(char) for char in data]
        self.writer.writerow(data)

    def write(self, data):
        if self.mode == "csv":
            self.write_csv(data)

    def __exit__(self, exc_type, exc_value, traceback):
        if self.writer:
            self.writer.close()
        self.file.close()



# -*- coding: utf-8 -*-
import platform

from setuptools import setup

import emokit

long_description = None

packages = [
    "emokit",
]

requirements = open("requirements.txt", "r").read().split("\n")
system_platform = platform.system()

if system_platform == "Windows":
    requirements += ['pywinusb']
else:
    requirements += ['pyhidapi']
setup(
    name="emokit",
    version=emokit.__version__,
    url="https://github.com/openyou/emokit",
    license="Public Domain, SEE LICENSE",
    author="Cody Brocious",
    author_email="cody.brocious@gmail.com",
    maintainer="Bill Schumacher",
    maintainer_email="bill@servernet.co",
    description="emotiv epoc eeg headset sdk",
    long_description=long_description,
    packages=packages,
    install_requires=requirements,
    scripts=[],
    platforms="any",
    zip_safe=False,
    classifiers=[
        "Operating System :: OS Independent",
        "Programming Language :: Python",
    ]
)



# -*- coding: utf-8 -*-

__all__ = ['battery', 'decrypter', 'sensors', 'emotiv', 'packet', 'util', 'reader']
"""
emokit - emotiv epoc reverse engineering
~~~~~~~~~~~~~~~

:copyright: Copyright (c) 2010-2012, Cody Brocious, Kyle Machulis/Nonpolynomial Labs, Copyright 2013 Bryan Bishop
    Copyright 2012-2016 Bill Schumacher/Cerebral Power
:license: Public Domain, See LICENSE
"""

__title__ = "emokit"
__version__ = "0.0.7"
__build__ = 0x000007
__author__ = "Cody Brocious / Kyle Machulis"
__license__ = "Public Domain, See LICENSE"
__copyright__ = "Copyright (c) 2010-2012, Cody Brocious, Kyle Machulis/Nonpolynomial Labs," \
                " Copyright 2013 Bryan Bishop, Copyright 2012-2016 Bill Schumacher/Cerebral Power"



# -*- coding: utf-8 -*-
import sys

from emokit.battery import battery_values
from emokit.sensors import sensor_bits, quality_bits, sensor_quality_bit, sensors_mapping
from emokit.util import get_level


class EmotivPacket(object):
    """
    Basic semantics for input bytes.
    """

    def __init__(self, data):
        """
        Initializes packet data. Sets the global battery value.
        Updates each sensor with current sensor value from the packet data.

        :param data - Values decrypted to be processed
        """
        if sys.version_info >= (3, 0):
            self.raw_data = [int(bit) for bit in data]
            data = self.raw_data
            self.counter = data[0]
        else:
            if type(data[0]) == str and len(data[0]) > 1:
                self.raw_data = [chr(int(bit)) for bit in data]
                data = self.raw_data
            else:
                self.raw_data = data
            self.counter = ord(data[0])
        self.battery = None
        if self.counter > 127:
            self.battery = battery_values[str(self.counter)]
            self.counter = 128
        self.sync = self.counter == 0xe9
        self.sensors = sensors_mapping.copy()
        if sys.version_info >= (3, 0):
            self.sensors['X']['value'] = data[29] - 106
            self.sensors['Y']['value'] = data[30] - 105
            self.sensors['Z']['value'] = '?'
        else:
            self.sensors['X']['value'] = ord(data[29]) - 106
            self.sensors['Y']['value'] = ord(data[30]) - 105
            self.sensors['Z']['value'] = '?'

        for name, bits in sensor_bits.items():
            # Get Level for sensors subtract 8192 to get signed value
            value = get_level(self.raw_data, bits) - 8192
            setattr(self, name, (value,))
            self.sensors[name]['value'] = value
        self.quality_bit, self.quality_value = self.handle_quality(self.sensors)

    def handle_quality(self, sensors):
        """
        Sets the quality value for the sensor from the quality bits in the packet data.
        Optionally will return the value.

        :param sensors - reference to sensors dict in Emotiv class.

        """
        current_contact_quality = get_level(self.raw_data, quality_bits)

        if sys.version_info >= (3, 0):
            sensor_bit = self.raw_data[0]
        else:
            sensor_bit = ord(self.raw_data[0])
        if sensor_quality_bit.get(sensor_bit, False):
            sensors[sensor_quality_bit[sensor_bit]]['quality'] = current_contact_quality
        else:
            sensors['Unknown']['quality'] = current_contact_quality
            sensors['Unknown']['value'] = sensor_bit
        return sensor_bit, current_contact_quality

    def __repr__(self):
        """
        Returns custom string representation of the Emotiv Packet.
        """
        return 'EmotivPacket(counter=%i, battery=%i, gyro_x=%i, gyro_y=%i, gyro_z=%s)' % (
            self.counter,
            self.battery,
            self.sensors['X']['value'],
            self.sensors['Y']['value'],
            self.sensors['Z']['value'])

    def get_quality_scale(self, old_model=False):
        if old_model:
            return self.quality_value // 540
        else:
            return self.quality_value // 1024
